# -*- coding: utf-8 -*-

from terraform_compliance.extensions.security_groups import SecurityGroup
from terraform_compliance.common.exceptions import TerraformComplianceInternalFailure
from terraform_compliance.common.error_handling import Error


def it_condition_have_proto_protocol_and_port_port_for_cidr(_step_obj, condition, proto, port, cidr):
    searching_for=dict(port=port, protocol=proto, cidr_blocks=cidr)

    for sg in _step_obj.context.stash:
        sg_obj = SecurityGroup(searching_for, sg['values'], address=sg['address'])
        if condition == 'must only':
            sg_obj.must_only_have()
        elif condition == 'must':
            sg_obj.must_have()
        elif condition == 'must not':
            sg_obj.must_not_have()
        else:
            raise TerraformComplianceInternalFailure('You can only use "must have", "must not have" and '
                                                     '"must only have" conditions on this step for now.'
                                                     'You tried to use "{}"'.format(condition))
        result, message = sg_obj.validate()

        if result is False:
            Error(_step_obj, message)

    return True


# -*- coding: utf-8 -*-


def i_flatten_everything_found(_step_obj):
    flattened_list = []
    addresses_flattened = []
    for each in _step_obj.context.stash:
        flattened_list.append(each.get('values', each))
        addresses_flattened.append(each.get('address', each))

    if flattened_list:
        _step_obj.context.stash = {
            'address': ', '.join(addresses_flattened),
            'values': flattened_list
        }

    return flattened_list



from unittest import TestCase
from terraform_compliance.common.terraform_files import which, convert_terraform_plan_to_json

import os
from mock import patch


class TestTerraformFiles(TestCase):

    def test_which_success(self):
        self.assertEqual('/bin/ls', which('/bin/ls'))

    def test_which_failure(self):
        self.assertIsNone(which('/something/that/doesnt/exist/hopefully'))

    @patch('terraform_compliance.common.terraform_files.which', return_value=None)
    def test_convert_terraform_plan_to_json(self, *args):
        with self.assertRaises(SystemExit):
            convert_terraform_plan_to_json('some_file')



import re
from netaddr import IPNetwork
from terraform_compliance.steps import resource_name
from collections.abc import Iterable
import json
from copy import deepcopy
import sys
from terraform_compliance.common.exceptions import TerraformComplianceInternalFailure
from semver import VersionInfo, compare
from terraform_compliance.common.defaults import Defaults
from radish.utils import console_write


class EmptyStash(object):
    pass


class Null(object):
    pass


class Match(object):
    def __init__(self, case_sensitive):
        self.case_sensitive = case_sensitive

    def equals(self, left, right):
        if not isinstance(left, (bool, int, float, str)) or not isinstance(right, (bool, int, float, str)):
            raise TypeError

        if self.case_sensitive:
            return str(left) == str(right)
        else:
            return str(left).lower() == str(right).lower()

    # gets something from a dictionary
    # needle == key
    def get(self, haystack, needle, default=None):
        if not isinstance(haystack, dict):
            return TypeError

        for key, value in haystack.items():
            if self.equals(key, needle):
                return value

        return default

    # checks if the collection (list, dict (key)) contains the element (only in first level, no seek)
    # needle should be a simple type
    # returns True for True in ['True']
    def contains(self, haystack, needle):
        if isinstance(haystack, (dict, list)):
            for key in haystack:
                if self.equals(key, needle):
                    return True
            return False

        raise TypeError

    # Assumes all incoming sets are sets of strings
    def issubset(self, set1, set2):
        if not isinstance(set1, Iterable) or isinstance(set1, str):
            raise TerraformComplianceInternalFailure('{} should be a non-str iterable'.format(set1))

        if not isinstance(set2, Iterable) or isinstance(set2, str):
            raise TerraformComplianceInternalFailure('{} should be a non-str iterable'.format(set2))

        if not self.case_sensitive:
            set1 = set(str(e).lower() for e in set1)
            set2 = set(str(e).lower() for e in set2)

        return set1 <= set2

    # re.match. overwrites the previous flag
    def regex_match(self, *args, **kwargs):
        regex_flag = 0 if self.case_sensitive else re.IGNORECASE
        return re.match(*args, **kwargs, flags=regex_flag)

    # seek key in dict but with case_sensitivity
    def seek_key_in_dict(self, haystack, needle):
        '''
        Searches needle in haystack ( could be dict, list, list of dicts, nested dicts, etc. ) and returns all findings
        as a list

        :param haystack: dict, list
        :param needle: search key
        :return: list of found keys & values
        '''
        found = list()
        if isinstance(haystack, dict):
            for key, value in haystack.items():
                if self.equals(key, needle):
                    found.append({key: value})
                else:
                    found.extend(self.seek_key_in_dict(value, needle))

        elif isinstance(haystack, list):
            for value in haystack:
                found.extend(self.seek_key_in_dict(value, needle))

        else:
            return []

        return found

    # ...
    # needle == value
    def seek_value_in_dict(self, needle, haystack, address=None):
        findings = []
        # if isinstance(haystack, (str, int, bool, float)) and str(needle) in str(haystack):  # this shouldn't be in but == instead
        #     findings.append(dict(values=needle, address=None))
        if isinstance(haystack, (str, int, bool, float)) and self.equals(needle, haystack):
            findings.append(dict(values=needle, address=None))

        elif isinstance(haystack, dict):
            address = haystack.get('address') if address is None else address

            for key, value in haystack.items():
                if isinstance(value, (dict, list)):
                    findings.extend(self.seek_value_in_dict(needle, value))

                elif isinstance(value, (str, bool, int, float)) and self.equals(needle, value):
                    findings.append(dict(values=needle, address=address))

        elif isinstance(haystack, list):
            # Check if this is a list of strings
            if all(isinstance(elem, str) for elem in haystack):
                findings.extend([elem for elem in haystack if self.equals(elem, needle)])

            # Otherwise, there are more stuff, so go recursive
            else:
                for value in haystack:
                    findings.extend(self.seek_value_in_dict(needle, value))

        return findings

    # ...
    # case sensitivity overwrites regex (if case insensitive, there is always re.IGNORECASE)
    # this's admittedly weird but convention/backwards compatibility...
    def seek_regex_key_in_dict_values(self, haystack, key_name, needle, key_matched=None):
        '''
        Searches needle in haystack ( could be dict, list, list of dicts, nested dicts, etc. ) and returns all findings
        as a list. The only difference from seek_key_in_dict is, we are assuming needle is in regex format here and we
        are searching for values instead.

        :param haystack: dict, list
        :param key_name: string of the key
        :param needle: regex search for the value
        :param key_matched: Internal use
        :return: list of found keys & values
        '''
        regex = r'^{}$'.format(needle)
        found = list()
        if isinstance(haystack, dict):
            for key, value in haystack.items():
                if isinstance(value, (bool, int, float)):
                    value = str(value)

                if self.equals(key, key_name) or key_matched is not None:
                    if isinstance(value, str):
                        matches = self.regex_match(regex, value)

                        if matches is not None:
                            found.append(matches.group(0))
                        else:
                            found.extend(self.seek_regex_key_in_dict_values(value, key_name, needle, True))

                    elif isinstance(value, dict):
                        found.extend(self.seek_regex_key_in_dict_values(value, key_name, needle, True))

                    elif isinstance(value, list):
                        for v in value:
                            found.extend(self.seek_regex_key_in_dict_values(v, key_name, needle, True))

                else:
                    found.extend(self.seek_regex_key_in_dict_values(value, key_name, needle, key_matched))

        elif isinstance(haystack, list):
            for value in haystack:
                found.extend(self.seek_regex_key_in_dict_values(value, key_name, needle, key_matched))

        else:
            return []

        return found


def flatten_list(input):
    return list(flatten(input))


def flatten(items):
    for x in items:
        if isinstance(x, Iterable) and not isinstance(x, (str, bytes)):
            for sub_x in flatten(x):
                yield sub_x
        else:
            yield x


def check_if_cidr(value):
    try:
        re.compile(value)
        return True
    except TypeError as e:
        return False


def is_ip_in_cidr(ip_cidr, cidr):
    is_ip_regex = r'(1[0-9][0-9]|2[0-4][0-9]|25[0-5]|[0-9][0-9]|[0-9])' \
                  r'\.' \
                  r'(1[0-9][0-9]|2[0-4][0-9]|25[0-5]|[0-9][0-9]|[0-9])' \
                  r'\.' \
                  r'(1[0-9][0-9]|2[0-4][0-9]|25[0-5]|[0-9][0-9]|[0-9])' \
                  r'\.' \
                  r'(1[0-9][0-9]|2[0-4][0-9]|25[0-5]|[0-9][0-9]|[0-9])' \
                  r'\/' \
                  r'(3[0-2]|2[0-9]|1[0-9]|[0-9])'

    # IP is a not a regex string
    if re.match(is_ip_regex, ip_cidr) is not None:
        for ip_network in cidr:
            if check_if_cidr(ip_cidr) and check_if_cidr(ip_network) and IPNetwork(ip_cidr) in IPNetwork(ip_network):
                return True

        return False

    # IP is a possible regex string
    else:
        # The given regex is matching the CIDR within the plan
        for elem in cidr:
            if re.match(ip_cidr, elem) is None:
                return False
        return True


def are_networks_same(first_network, network_list):
    for second_network in network_list:
        if check_if_cidr(first_network) and check_if_cidr(second_network) and \
                IPNetwork(first_network) == IPNetwork(second_network):
            return True

    return False


def convert_resource_type(resource_type):
    '''
    Searchs given resource_type within resource_name array and returns the value if it is found

    :param resource_type: String of resource_type
    :return: converted or original resource_type
    '''
    if resource_type in resource_name.keys():
        resource_type = resource_name[resource_type]

    return resource_type


def seek_key_in_dict(haystack, needle):
    '''
    Searches needle in haystack ( could be dict, list, list of dicts, nested dicts, etc. ) and returns all findings
    as a list

    :param haystack: dict, list
    :param needle: search key
    :return: list of found keys & values
    '''
    found = list()
    if isinstance(haystack, dict):
        for key, value in haystack.items():
            if key.lower() == needle.lower():
                found.append({key: value})
            else:
                found.extend(seek_key_in_dict(value, needle))

    elif isinstance(haystack, list):
        for value in haystack:
            found.extend(seek_key_in_dict(value, needle))

    else:
        return []

    return found


def seek_regex_key_in_dict_values(haystack, key_name, needle, key_matched=None):
    '''
    Searches needle in haystack ( could be dict, list, list of dicts, nested dicts, etc. ) and returns all findings
    as a list. The only difference from seek_key_in_dict is, we are assuming needle is in regex format here and we
    are searching for values instead.

    :param haystack: dict, list
    :param key_name: string of the key
    :param needle: regex search for the value
    :param key_matched: Internal use
    :return: list of found keys & values
    '''
    regex = r'^{}$'.format(needle)
    found = list()
    if isinstance(haystack, dict):
        for key, value in haystack.items():
            if isinstance(value, (bool, int, float)):
                value = str(value)

            if key.lower() == key_name.lower() or key_matched is not None:
                if isinstance(value, str):
                    matches = re.match(regex, value, flags=re.IGNORECASE)

                    if matches is not None:
                        found.append(matches.group(0))
                    else:
                        found.extend(seek_regex_key_in_dict_values(value, key_name, needle, True))

                elif isinstance(value, dict):
                    found.extend(seek_regex_key_in_dict_values(value, key_name, needle, True))

                elif isinstance(value, list):
                    for v in value:
                        found.extend(seek_regex_key_in_dict_values(v, key_name, needle, True))

            else:
                found.extend(seek_regex_key_in_dict_values(value, key_name, needle, key_matched))

    elif isinstance(haystack, list):
        for value in haystack:
            found.extend(seek_regex_key_in_dict_values(value, key_name, needle, key_matched))

    else:
        return []

    return found


def find_root_by_key(haystack, needle, return_key=None, _inherited_key=None, _depth=0, _return_value=None):
    '''
    Searches needle in haystack ( could be dict, list, list of dicts, nested dicts, etc. ) and returns the root key
    that has this needle somewhere within it's children.

    :param haystack: dict, list
    :param needle: search key
    :param return_key: if this is given, then the result will be the root_key[return_key] instead of root_key
    :param _inherited_key: internal usage, do not pass this.
    :param _depth: internal usage, do not pass this.
    :param _return_value: internal usage, do not pass this.
    :return:
    '''
    found = list()
    if isinstance(haystack, dict):
        for key, value in haystack.items():
            if not _depth:
                _inherited_key = key
                _return_value = key if not return_key else haystack[_inherited_key].get(return_key, _inherited_key)

            if key.lower() == needle.lower():
                found.append(_return_value)
            else:
                found.extend(find_root_by_key(value, needle, return_key, _inherited_key, _depth + 1, _return_value))

    elif isinstance(haystack, list) and _inherited_key is not None:
        for value in haystack:
            found.extend(find_root_by_key(value, needle, return_key, _inherited_key, _depth + 1, _return_value))

    else:
        return []

    return list(set(found))


def jsonify(string):
    if not isinstance(string, str):
        return string

    try:
        return json.loads(string)
    except json.decoder.JSONDecodeError:
        return string


def recursive_jsonify(haystack):
    if isinstance(haystack, str):
        haystack = jsonify(haystack)
        if isinstance(haystack, (list, dict)):
            return recursive_jsonify(haystack)
        return haystack

    if isinstance(haystack, dict):
        haystack = {key: recursive_jsonify(value) for key, value in haystack.items()}

    if isinstance(haystack, list):
        haystack = [recursive_jsonify(value) for value in haystack]

    return haystack


def get_resource_name_from_stash(stash, alternative_stash=None, address=None):
    if address is not None:
        return {'address': address}

    if isinstance(alternative_stash, (str, bool)) or alternative_stash is None:
        if isinstance(stash, list):

            # Get the first number, since this is usually due to `count` usage in terraform
            if 'address' in stash[0] and stash[0]['address'] is not None:
                return {'address': stash[0]['address'].replace('[0]', '')}
            else:
                return {'address': stash[0]}

        else:
            if 'address' in alternative_stash:
                return alternative_stash
            else:
                return {'address': alternative_stash}


def get_resource_address_list_from_stash(resource_list):
    address_list = []
    for resource in resource_list:
        if 'address' in resource and resource['address'] not in address_list:
            address_list.append(resource['address'])

    return address_list


def remove_mounted_resources(resource_list):
    if not isinstance(resource_list, list):
        return resource_list

    resources = deepcopy(resource_list)
    for resource in resources:
        if 'terraform-compliance.mounted_resources' in resource:
            for key, mounted_resource_type in resource['terraform-compliance.mounted_resources'].items():
                if mounted_resource_type in resource['values']:
                    del resource['values'][mounted_resource_type]

    return resources


def search_regex_in_list(regex, target_list):
    if isinstance(target_list, list):
        return list(filter(re.compile(r'{}'.format(regex)).match, target_list))

    return False


def seek_value_in_dict(needle, haystack, address=None):
    findings = []
    if isinstance(haystack, (str, int, bool, float)) and str(needle) in str(haystack):
        findings.append(dict(values=needle, address=None))

    elif isinstance(haystack, dict):
        address = haystack.get('address') if address is None else address

        for key, value in haystack.items():
            if isinstance(value, (dict, list)):
                findings.extend(seek_value_in_dict(needle, value))
            elif isinstance(value, (str, bool, int, float)) and needle.lower() == str(value).lower():
                findings.append(dict(values=needle, address=address))

    elif isinstance(haystack, list):
        # Check if this is a list of strings
        if all(isinstance(elem, str) for elem in haystack):
            findings.extend([elem for elem in haystack if elem.lower() == needle.lower()])

        # Otherwise, there are more stuff, so go recursive
        else:
            for value in haystack:
                findings.extend(seek_value_in_dict(needle, value))

    return findings


def dict_merge(source, target):
    if not isinstance(source, dict) or not isinstance(target, dict):
        return source

    src = deepcopy(source)
    dst = deepcopy(target)
    for key, val in src.items():
        if key in dst and is_list_of_dict(val) and is_list_of_dict(dst[key]):
            for dst_elem in dst[key]:
                for each_key in dst_elem.keys():
                    if not is_key_exist(each_key, val):
                        src[key].append(dst_elem)

    return src


def is_list_of_dict(target_list):
    return isinstance(target_list, list) and len(target_list) and isinstance(target_list[0], dict)


def is_key_exist(key, target_list_of_dict):
    for each in target_list_of_dict:
        if isinstance(each, dict) and key in each.keys():
            return True

    return False


def transform_asg_style_tags(resource_list):
    if not isinstance(resource_list, list):
        return resource_list

    for resource in resource_list:
        tag = resource.get('values', {}).get('tag')

        if not resource.get('values', {}).get('tags'):
            resource['values']['tags'] = {}

        if tag:
            for elem in tag:
                if 'key' in elem and 'value' in elem:
                    resource['values']['tags'][elem['key']] = elem['value']

    return resource_list


def python_version_check():
    python_version = sys.version.split(' ')[0]

    if not python_version:
        raise TerraformComplianceInternalFailure('Could not determine python version. '
                                                 'Please post this to issues: '.format(sys.version))

    python_version = VersionInfo.parse(python_version)

    if compare(str(python_version), Defaults.supported_min_python_versions) < 0:
        console_write('ERROR: Python version {} is not supported. '
                      'You must have minimum {} version.'.format(python_version,
                                                                 Defaults.supported_min_python_versions[0]))
        sys.exit(1)

    return True

def merge_dicts(source, target):
    '''
    successor of dict_merge
    merges two dictionaries or lists into one
    overlapping list/dictionaries are also merged within each other
   
    assume target is low priority

    conflicts
        regular values: higher prio overrides
        dictionaries: merge conflicting dictionaries
        lists: higher prio overrides for regular values, merge any matching dictionary/list elements
            if len(target) > len(source):
                append the tail to the source


    # Confusing and less important comments:
    if isinstance(source, list) and isinstance(target, list):
        # shady, just extend?
        for i, v in enumerate(higher prio):
            if v is dict
                if lower has dict in the ith element
                    combine them
    if isinstance(source, dict) and isinstance(target, dict):
        if all elements are singular
            return {**target, **source}
        merged = {**target, **source} # source has priority over conflicts
        # find conflicting lists
        for all conflicting lists,
            # merged[key] = source[key] + target[key] # honestly depends on the situation
            merged[key] = merge_dicts(source[key], target[key])

        for all conflicting lists
            merged[key] = merge_dicts(source[key], target[key])
    decisions to make
        what happens to conflicting lists?
            should I drop one and keep the other (could be relevant on some cases)
            should I merge the lists? (could be relevant as well)

        what happens to conflicting dirs?
    if I have two lists
        lists have dictionaries in the same index
            merge those dictionaries

        # add remaining values of lower priority to higher priority
        ignore non dict values
    '''
    if isinstance(source, list) and isinstance(target, list):
        for i, value in enumerate(target):
            if i >= len(source):
                source.append(value)
            elif isinstance(value, dict) and isinstance(source[i], dict):
                merge_dicts(source[i], target[i])
            elif isinstance(value, list) and isinstance(source[i], list):
                merge_dicts(source[i], target[i])

    if isinstance(source, dict) and isinstance(target, dict):
        for key, value in target.items():
            if key not in source:
                source[key] = target[key]
            else:
                # regular values
                # do nothing
                # dictionaries
                if isinstance(value, dict) and isinstance(source[key], dict):
                    merge_dicts(source[key], target[key])
                elif isinstance(value, list) and isinstance(source[key], list):
                    merge_dicts(source[key], target[key])


def remove_constant_values(target):
    if isinstance(target, dict):
        for key, value in target.items():
            if isinstance(value, dict) and len(value) == 1 and 'constant_value' in value:
                target[key] = value['constant_value']
            elif isinstance(value, dict) or isinstance(value, list):
                remove_constant_values(value)
    elif isinstance(target, list):
        for value in target:
            remove_constant_values(value)


import json
import pytest
from pathlib import Path
from terraform_compliance.steps.steps import (
    its_key_is_value, 
    its_key_is_not_value,
)
from tests.mocks import MockedStep, MockedWorld
from mock import patch

class ItsKeyValueTestCase:
    pass

def _get_test_cases():
    def build_test_case(p):
        with p.open() as f:
            j = json.load(f)
            test_case = ItsKeyValueTestCase()
            test_case.name = p.stem
            test_case.stash = j['stash']
            test_case.args = j['args']
            test_case.expected_ids = set(j['expected_ids'])
            return test_case

    path = Path(__file__).parent / 'its_key_value_test_cases'
    for p in path.rglob('*.json'):
        yield build_test_case(p)

class TestItsKeyValueStepCases():
    @pytest.mark.parametrize("test_case", list(_get_test_cases()), ids=lambda x: x.name)
    def test_its_key_is_value(self, test_case):
        step = MockedStep()
        step.context.stash = test_case.stash

        with patch('terraform_compliance.extensions.ext_radish_bdd.world', return_value=MockedWorld()):
            its_key_is_value(step, *test_case.args)

        assert step.state == ('passed' if test_case.expected_ids else 'skipped')

        if test_case.expected_ids:
            actual_ids = set(o['id'] for o in step.context.stash)
            assert actual_ids == test_case.expected_ids

    @pytest.mark.skip()
    @pytest.mark.parametrize("test_case", list(_get_test_cases()), ids=lambda x: x.name)
    def test_its_key_is_not_value(self, test_case):
        step = MockedStep()
        step.context.stash = test_case.stash

        with patch('terraform_compliance.extensions.ext_radish_bdd.world', return_value=MockedWorld()):
            its_key_is_not_value(step, *test_case.args)

        def collect_ids(val):
            if isinstance(val, dict):
                for k, v in val.items():
                    if k == 'id':
                        yield v
                    else: 
                        yield from collect_ids(v)
            elif isinstance(val, list):
                for child in val:
                    yield from collect_ids(child)

        all_ids = set(collect_ids(test_case.stash))
        expected_ids = all_ids - test_case.expected_ids

        assert step.state == ('passed' if expected_ids else 'skipped')
        
        if expected_ids:
            actual_ids = set(o['id'] for o in step.context.stash)
            assert actual_ids == expected_ids



from unittest import TestCase
from terraform_compliance.extensions.terraform import TerraformParser, seek_key_in_dict
from tests.mocks import MockedData
from mock import patch
from ddt import ddt, data
from copy import deepcopy


@ddt
class TestTerraformParser(TestCase):


    @patch.object(TerraformParser, '_read_file', return_value={})
    def test_version_check_success(self, *args):
        obj = TerraformParser('somefile', parse_it=False)
        obj.raw['format_version'] = obj.supported_format_versions[0]
        obj.raw['terraform_version'] = obj.supported_terraform_versions[0]
        self.assertTrue(obj._version_check())


    @patch.object(TerraformParser, '_read_file', return_value={})
    def test_version_check_failure_unsupported_format_version(self, *args):
        obj = TerraformParser('somefile', parse_it=False)
        obj.raw['format_version'] = 'some_unsupported_format_version'
        obj.raw['terraform_version'] = obj.supported_terraform_versions[0]
        with self.assertRaises(SystemExit):
            obj._version_check()


    @patch.object(TerraformParser, '_read_file', return_value={})
    def test_version_check_failure_unsupported_terraform_version(self, *args):
        obj = TerraformParser('somefile', parse_it=False)
        obj.raw['format_version'] = obj.supported_format_versions[0]
        obj.raw['terraform_version'] = 'some_unsupported_terraform_version'
        with self.assertRaises(SystemExit):
            obj._version_check()


    @patch.object(TerraformParser, '_read_file', return_value={})
    def test_identify_data_file_success(self, *args):
        obj = TerraformParser('somefile', parse_it=False)
        obj.raw['values'] = True
        obj._identify_data_file()
        self.assertEqual(obj.file_type, 'state')


    @patch.object(TerraformParser, '_read_file', return_value={})
    def test_identify_data_file_failure(self, *args):
        obj = TerraformParser('somefile', parse_it=False)
        obj._identify_data_file()
        self.assertEqual(obj.file_type, 'plan')


    @patch.object(TerraformParser, '_read_file', return_value={})
    def test_parse_variables(self, *args):
        obj = TerraformParser('somefile', parse_it=False)
        obj.raw['variables'] = True
        obj._parse_variables()
        self.assertEqual(obj.variables, True)


    @patch.object(TerraformParser, '_read_file', return_value={})
    def test_parse_resources_resources_exists_in_the_plan_data(self, *args):
        obj = TerraformParser('somefile', parse_it=False)
        obj.raw['planned_values'] = {
            'root_module': {
                'resources': [
                    {
                        'address': 'data.something'
                    }
                ]
            }
        }
        obj._parse_resources()
        self.assertEqual(obj.data['data.something'], {'address': 'data.something'})


    @patch.object(TerraformParser, '_read_file', return_value={})
    def test_parse_resources_resources_exists_in_the_plan_resource(self, *args):
        obj = TerraformParser('somefile', parse_it=False)
        obj.raw['planned_values'] = {
            'root_module': {
                'resources': [
                    {
                        'address': 'something'
                    }
                ]
            }
        }
        obj._parse_resources()
        self.assertEqual(obj.resources['something'], {'address': 'something'})


    @patch.object(TerraformParser, '_read_file', return_value={})
    def test_parse_resources_resources_exists_in_the_state_data(self, *args):
        obj = TerraformParser('somefile', parse_it=False)
        obj.raw['values'] = {
            'root_module': {
                'resources': [
                    {
                        'address': 'data.something'
                    }
                ]
            }
        }
        obj._parse_resources()
        self.assertEqual(obj.data['data.something'], {'address': 'data.something'})


    @patch.object(TerraformParser, '_read_file', return_value={})
    def test_parse_resources_resources_exists_in_the_state_resource(self, *args):
        obj = TerraformParser('somefile', parse_it=False)
        obj.raw['values'] = {
            'root_module': {
                'resources': [
                    {
                        'address': 'something'
                    }
                ]
            }
        }
        obj._parse_resources()
        self.assertEqual(obj.resources['something'], {'address': 'something'})


    @patch.object(TerraformParser, '_read_file', return_value={})
    def test_parse_resources_child_resources_exists_in_the_state_data(self, *args):
        obj = TerraformParser('somefile', parse_it=False)
        obj.raw['values'] = {
            'root_module': {
                'child_modules': {
                    'resources': [
                        {
                            'address': 'data.something'
                        }
                    ]
                }
            }
        }
        obj._parse_resources()
        self.assertEqual(obj.data['data.something'], {'address': 'data.something'})


    @patch.object(TerraformParser, '_read_file', return_value={})
    def test_parse_resources_child_resources_exists_in_the_state_resource(self, *args):
        obj = TerraformParser('somefile', parse_it=False)
        obj.raw['values'] = {
            'root_module': {
                'child_modules': {
                    'resources': [
                        {
                            'address': 'something'
                        }
                    ]
                }
            }
        }
        obj._parse_resources()
        self.assertEqual(obj.resources['something'], {'address': 'something'})

    @patch.object(TerraformParser, '_read_file', return_value={})
    def test_parse_resources_resources_exists_in_the_resource_changes_data(self, *args):
        obj = TerraformParser('somefile', parse_it=False)
        obj.raw['resource_changes'] = [
            {
                'mode': 'data',
                'address': 'data.something',
                'change': {
                    'actions': ['update'],
                    'before': {
                        'key': 'foo'
                    },
                    'after': {
                        'key': 'bar'
                    }
                }
            }
        ]
        obj._parse_resources()
        self.assertEqual(obj.data['data.something'], {'address': 'data.something', 'mode': 'data', 'values': {'key': 'bar'}})

    @patch.object(TerraformParser, '_read_file', return_value={})
    def test_parse_resources_resources_exists_in_the_resource_changes_resource(self, *args):
        obj = TerraformParser('somefile', parse_it=False)
        obj.raw['resource_changes'] = [
            {
                'address': 'something',
                'change': {
                    'actions': ['update'],
                    'before': {
                        'key': 'foo'
                    },
                    'after': {
                        'key': 'bar'
                    }
                }
            }
        ]
        obj._parse_resources()
        self.assertEqual(obj.resources['something'], {'address': 'something', 'values': {'key': 'bar'}})

    @patch.object(TerraformParser, '_read_file', return_value={})
    def test_parse_resources_resources_exists_in_the_resource_changes_deleted(self, *args):
        obj = TerraformParser('somefile', parse_it=False)
        obj.raw['resource_changes'] = [
            {
                'address': 'something',
                'change': {
                    'actions': ['delete'],
                    'before': {
                        'key': 'foo'
                    },
                    'after': {
                        'key': 'bar'
                    }
                }
            }
        ]
        obj._parse_resources()
        self.assertEqual(obj.resources, {})

    @patch.object(TerraformParser, '_read_file', return_value={})
    def test_parse_configurations_resources(self, *args):
        obj = TerraformParser('somefile', parse_it=False)
        obj.raw['configuration'] = {
            'root_module': {
                'resources': [
                    {
                        'address': 'something'
                    }
                ]
            }
        }
        obj._parse_configurations()
        self.assertEqual(obj.configuration['resources']['something'], {'address': 'something'})


    @patch.object(TerraformParser, '_read_file', return_value={})
    def test_parse_configurations_variables(self, *args):
        obj = TerraformParser('somefile', parse_it=False)
        obj.raw['configuration'] = {
            'root_module': {
                'variables': [
                    {
                        'variable_key': 'variable_value'
                    }
                ]
            }
        }
        obj._parse_configurations()
        self.assertEqual(obj.configuration['variables'], [{'variable_key': 'variable_value'}])


    @patch.object(TerraformParser, '_read_file', return_value={})
    def test_parse_configurations_providers(self, *args):
        obj = TerraformParser('somefile', parse_it=False)
        obj.raw['configuration'] = {
            'provider_config': {
                'some_provider': 'some_provider_data'
            }
        }
        obj._parse_configurations()
        self.assertEqual(obj.configuration['providers'], {'some_provider': 'some_provider_data'})

    @patch.object(TerraformParser, '_read_file', return_value={})
    def test_parse_configurations_module_resources(self, *args):
        obj = TerraformParser('somefile', parse_it=False)
        obj.raw['configuration'] = {
            'root_module': {
                'module_calls': {
                    'a_module':{
                        'module': {
                            'resources': [
                                {
                                    'address': 'something'
                                }
                            ]
                        }
                    }
                }
            }
        }
        obj._parse_configurations()
        self.assertEqual(obj.configuration['resources']['something'], {'address': 'something'})

    @patch.object(TerraformParser, '_read_file', return_value={})
    def test_parse_configurations_ignore_nested_resources(self, *args):
        obj = TerraformParser('somefile', parse_it=False)
        obj.raw['configuration'] = {
            'root_module': {
                'module_calls': {
                    'a_module':{
                        'expressions': {
                            'cluster_role_rules': {
                                'constant_values': [
                                    {
                                        "resources": [
                                            "namespaces",
                                            "pods"
                                        ]
                                    }
                                ]
                            }
                        }
                    }
                },
                'resources': [
                    {
                        'address': 'something'
                    }
                ]
            }
        }
        obj._parse_configurations()
        self.assertEqual(obj.configuration['resources']['something'], {'address': 'something'})

    @patch.object(TerraformParser, '_read_file', return_value={})
    def test_mount_resources_with_ref_type(self, *args):
        obj = TerraformParser('somefile', parse_it=False)
        target = 'target_ref'
        source = 'source_ref'
        ref_type = 'some_key'
        obj.resources = {
            target: {
                'values': {
                    'some_key': ['a']
                }
            },
            source: {
                'values': {
                    'some_key': 'some_value'
                }
            }
        }
        obj.resources_raw = deepcopy(obj.resources)
        obj._mount_resources([source], {'some_key': [target]}, ref_type)
        self.assertEqual(['a', {'some_key': 'some_value', 'terraform-compliance.mounted': True}], obj.resources[target]['values'][ref_type])

    @patch.object(TerraformParser, '_read_file', return_value={})
    def test_mount_resources_without_ref_type(self, *args):
        obj = TerraformParser('somefile', parse_it=False)
        target = 'target_ref'
        source = 'source_ref'
        ref_type = 'some_ref'
        obj.resources = {
            target: {
                'values': {}
            },
            source: {
                'values': {
                    'some_key': 'some_value'
                }
            }
        }
        obj.resources_raw = deepcopy(obj.resources)
        obj._mount_resources([source], {'some_key': [target]}, ref_type)
        self.assertEqual([{'some_key': 'some_value', 'terraform-compliance.mounted': True}], obj.resources[target]['values'][ref_type])

    @patch.object(TerraformParser, '_read_file', return_value={})
    def test_mount_resources_failure_if_source_resource_does_not_have_values(self, *args):
        obj = TerraformParser('somefile', parse_it=False)
        target = 'target_ref'
        source = 'source_ref'
        ref_type = 'some_ref'
        obj.resources = {
            target: {
                'values': {}
            },
            source: {
            }
        }
        obj.resources_raw = deepcopy(obj.resources)
        obj._mount_resources([source], [target], ref_type)
        self.assertEqual({}, obj.resources[target]['values'])


    @patch.object(TerraformParser, '_read_file', return_value={})
    def test_find_resource_from_name_resource_name_in_resources(self, *args):
        obj = TerraformParser('somefile', parse_it=False)
        obj.resources = {'something': True}
        self.assertEqual(obj._find_resource_from_name('something'), ['something'])


    @patch.object(TerraformParser, '_read_file', return_value={})
    def test_find_resource_from_name_resource_name_in_child_resources(self, *args):
        obj = TerraformParser('somefile', parse_it=False)
        obj.resources = {
            'a': {
                'type': 'resource_type',
                'name': 'resource_id'
            }
        }
        self.assertEqual(obj._find_resource_from_name('resource_type.resource_id'), ['a'])

    @patch.object(TerraformParser, '_read_file', return_value={})
    def test_find_resource_from_name_resource_name_in_module_outputs_expression_references(self, *args):
        obj = TerraformParser('somefile', parse_it=False)
        obj.raw['configuration'] = {
            "root_module": {
            "module_calls": {
                "module_name": {
                    "source": "../../module-test",
                    "module": {
                        "outputs": {
                            "my_expression": {
                                "expression": {
                                    "references": [
                                        "var.example"
                                    ]
                                }
                            }
                        }
                    }
                }
            }
            }
        }
        self.assertEqual(obj._find_resource_from_name('module.module_name.my_expression'), ['module.module_name.var.example'])

    @patch.object(TerraformParser, '_read_file', return_value={})
    def test_find_resource_from_name_resource_name_in_module_outputs_expression(self, *args):
        obj = TerraformParser('somefile', parse_it=False)
        obj.raw['configuration'] = {
            "root_module": {
            "module_calls": {
                "module_name": {
                    "source": "../../module-test",
                    "module": {
                        "outputs": {
                            "my_expression": {
                                "expression": {
                                    "constant_value": [
                                        "10.123.0.0/16"
                                    ]
                                },
                                "description": "List of cidr(s)"
                            }
                        }
                    }
                }
            }
            }
        }
        self.assertEqual(obj._find_resource_from_name('module.module_name.my_expression'), [])

    @patch.object(TerraformParser, '_read_file', return_value={})
    def test_distribute_providers(self, *args):
        obj = TerraformParser('somefile', parse_it=False)
        obj.resources = {
            'provider_resource_id': {
                'type': 'resource_type',
                'name': 'resource_id'
            },
            'other-provider_other_resource_id': {
                'type': 'other-provider_other_resource_type',
                'name': 'resource_id'
            }
        }
        obj._distribute_providers()

        print(obj.providers)
        self.assertTrue('provider' in obj.providers)
        self.assertTrue('other-provider' in obj.providers)

    @patch.object(TerraformParser, '_read_file', return_value={})
    @patch.object(TerraformParser, '_version_check', return_value=None)
    @patch.object(TerraformParser, '_identify_data_file', return_value=None)
    @patch.object(TerraformParser, '_parse_resources', return_value=None)
    @patch.object(TerraformParser, '_parse_configurations', return_value=None)
    @patch.object(TerraformParser, '_mount_references', return_value=None)
    @patch.object(TerraformParser, '_distribute_providers', return_value=None)
    def test_parse_plan(self, *args):
        obj = TerraformParser('somefile', parse_it=False)
        obj.raw['variables'] = True
        obj.parse()

        self.assertEqual(obj.variables, True)

    @patch.object(TerraformParser, '_read_file', return_value={})
    @patch.object(TerraformParser, '_version_check', return_value=None)
    @patch.object(TerraformParser, '_identify_data_file', return_value=None)
    @patch.object(TerraformParser, '_parse_resources', return_value=None)
    @patch.object(TerraformParser, '_parse_configurations', return_value=None)
    @patch.object(TerraformParser, '_mount_references', return_value=None)
    @patch.object(TerraformParser, '_distribute_providers', return_value=None)
    def test_parse_state(self, *args):
        obj = TerraformParser('somefile', parse_it=False)
        obj.raw['variables'] = True
        obj.file_type = 'state'
        obj.parse()

        self.assertEqual(obj.variables, None)


    @patch.object(TerraformParser, '_read_file', return_value={})
    def test_find_resources_by_type_success(self, *args):
        obj = TerraformParser('somefile', parse_it=False)
        obj.resources = {
            'provider_resource_id': {
                'type': 'resource_type',
                'name': 'resource_id',
                'mode': 'managed'
            },
            'other-provider_other_resource_id': {
                'type': 'other-provider_other_resource_type',
                'name': 'resource_id',
                'mode': 'managed'
            }
        }

        self.assertEqual(obj.find_resources_by_type('resource_type'),
                         [{
                            'type': 'resource_type',
                            'name': 'resource_id',
                            'mode': 'managed'
                          }])

    @patch.object(TerraformParser, '_read_file', return_value={})
    def test_find_resources_by_type_failure(self, *args):
        obj = TerraformParser('somefile', parse_it=False)
        obj.resources = {
            'provider_resource_id': {
                'type': 'resource_type',
                'name': 'resource_id',
                'mode': 'managed'
            },
            'other-provider_other_resource_id': {
                'type': 'other-provider_other_resource_type',
                'name': 'resource_id',
                'mode': 'managed'
            }
        }

        self.assertEqual(obj.find_resources_by_type('invalid_resource'),
                         [])

    @patch.object(TerraformParser, '_read_file', return_value={})
    def test_expand_resource_tags_success(self, *args):
        obj = TerraformParser('somefile', parse_it=False)
        obj.resources = {
            'provider_resource_id': {
                'type': 'resource_type',
                'name': 'resource_id',
                'mode': 'managed',
                'values': {
                    'tags': [
                        {
                        'key': 'somekey',
                        'value': 'somevalue',
                        'someotherkey': True
                        }
                    ]
                }
            },
            'other-provider_other_resource_id': {
                'type': 'other-provider_other_resource_type',
                'name': 'resource_id',
                'mode': 'managed',
                'values': {}
            }
        }
        obj._expand_resource_tags(obj.resources['provider_resource_id'])
        obj._expand_resource_tags(obj.resources['other-provider_other_resource_id'])
        self.assertTrue('somekey' in obj.resources['provider_resource_id']['values']['tags'][0])
        self.assertEqual(obj.resources['provider_resource_id']['values']['tags'][0]['somekey'], 'somevalue')
        self.assertEqual(obj.resources['provider_resource_id']['values']['tags'][0]['key'], 'somekey')
        self.assertEqual(obj.resources['provider_resource_id']['values']['tags'][0]['someotherkey'], True)
        self.assertTrue('tags' not in obj.resources['other-provider_other_resource_id']['values'])

    @patch.object(TerraformParser, '_read_file', return_value={})
    @data([], {}, '', 123, [dict(a='b')])
    def test_expand_resource_tags_failures(self, value, *args):
        obj = TerraformParser('somefile', parse_it=False)
        obj.resources = {
            'provider_resource_id': {
                'type': 'resource_type',
                'name': 'resource_id',
                'mode': 'managed',
                'values': {
                    'tags': [
                        value
                    ]
                }
            },
            'other-provider_other_resource_id': {
                'type': 'other-provider_other_resource_type',
                'name': 'resource_id',
                'mode': 'managed',
                'values': {}
            }
        }
        obj._expand_resource_tags(obj.resources['provider_resource_id'])
        obj._expand_resource_tags(obj.resources['other-provider_other_resource_id'])
        self.assertTrue('tags' not in obj.resources['other-provider_other_resource_id']['values'])
        self.assertEqual(value, obj.resources['provider_resource_id']['values']['tags'][0])

    @patch.object(TerraformParser, '_read_file', return_value={})
    def test_remember_after_unknown_simple(self, *args):
        obj = TerraformParser('somefile', parse_it=False)
        resource = {
            'type': 'resource_type',
        }

        after_unknown_values = {'property_{}'.format(i): i for i in range(10)}
        obj.remember_after_unknown(resource, after_unknown_values)

        self.assertTrue(resource['type'] in obj.type_to_after_unknown_properties)

        for key in after_unknown_values:
            mapped_after_unknown_value = obj.type_to_after_unknown_properties[resource['type']][key]
            self.assertEqual(mapped_after_unknown_value, after_unknown_values[key])
    
    @patch.object(TerraformParser, '_read_file', return_value={})
    def test_remember_after_unknown(self, *args):
        obj = TerraformParser('somefile', parse_it=False)

        # test parameters (not all combinations will cover all properties for all types)
        len_resources = 20
        len_properties = 20
        len_types = 3

        # create resources that falls into specific types
        resources = {
            'resource_{}'.format(i): {'type': 'type_{}'.format(i % len_types)} for i in range(len_resources)
        }
        
        # create after_unknown for each.
        # resource will have varying after_unknown values despite belonging the same type
        resource_to_after_unknown_values = {}
        for j, resource in enumerate(resources):
            if j < 3:
                resource_to_after_unknown_values[resource] = {}
            else:
                resource_to_after_unknown_values[resource] = {'property_{}'.format(i): i for i in range(j % 2, len_properties, 2)}

        # check if all after_unknown values are accumulated correctly per each resource type
        for resource_name, resource in resources.items():
            obj.remember_after_unknown(resource, resource_to_after_unknown_values[resource_name])


        for i in range(len_types):
            t = 'type_{}'.format(i)

            self.assertTrue(t in obj.type_to_after_unknown_properties)

            for property_value in range(len_properties):
                property_key = 'property_{}'.format(property_value)

                mapped_after_unknown_value = obj.type_to_after_unknown_properties[t][property_key]
                self.assertEqual(mapped_after_unknown_value, property_value)                



# -*- coding: utf-8 -*-

from terraform_compliance.common.helper import (
    search_regex_in_list
)
from terraform_compliance.common.error_handling import Error
from terraform_compliance.common.exceptions import TerraformComplianceInternalFailure
from terraform_compliance.common.defaults import Defaults


def it_must_have_reference_address_referenced(_step_obj, reference_address):
    if _step_obj.context.stash:
        for resource in _step_obj.context.stash:
            if isinstance(resource, dict):
                if Defaults.r_mount_addr_ptr_list in resource and search_regex_in_list(reference_address,
                                                                                       resource[Defaults.r_mount_addr_ptr_list]):
                    return True
            else:
                raise TerraformComplianceInternalFailure('Unexpected resource structure: {}'.format(resource))

            Error(_step_obj, '{} is not referenced within {}.'.format(reference_address, resource.get('address')))
    else:
        Error(_step_obj, 'No entities found for this step to process. Check your filtering steps in this scenario.')



# -*- coding: utf-8 -*-

from terraform_compliance.common.defaults import Defaults
from terraform_compliance.common.helper import (
    EmptyStash,
    Null
)
from terraform_compliance.common.exceptions import TerraformComplianceNotImplemented
from terraform_compliance.common.error_handling import Error


def i_action_them(_step_obj, action_type):
    match = _step_obj.context.match
    seek_key_in_dict = match.seek_key_in_dict

    if action_type == "count":
        # WARNING: Only case where we set stash as a dictionary, instead of a list.
        if isinstance(_step_obj.context.stash, list):

            # This means we are directly started counting without drilling down any property
            # Thus, our target for the count is stash itself.
            if match.contains(Defaults().types_list, _step_obj.context.property_name):
                _step_obj.context.stash = dict(values=len(_step_obj.context.stash))

            else:
                if isinstance(_step_obj.context.stash[0], dict):
                    if _step_obj.context.stash[0].get('values'):
                        _step_obj.context.stash = seek_key_in_dict(_step_obj.context.stash, 'values')
                        count = 0
                        for result in _step_obj.context.stash:
                            count += len(result.get('values', {})) if result.get('values') and not isinstance(result.get('values'), (int,bool,str)) else 1

                        _step_obj.context.stash = dict(values=count)

                else:
                    _step_obj.context.stash = dict(values=len(_step_obj.context.stash))
    else:
        raise TerraformComplianceNotImplemented('Invalid action_type in the scenario: {}'.format(action_type))


def i_expect_the_result_is_operator_than_number(_step_obj, operator, number, _stash=EmptyStash):

    def fail(assertion, message):
        try:
            assert assertion, 'Failed'
        except AssertionError as e:
            Error(_step_obj, 'for {} on {}. {}.'.format(_step_obj.context.address,
                                                        _step_obj.context.property_name,
                                                        message))

    values = _step_obj.context.stash if _stash is EmptyStash else _stash

    if isinstance(values, list):
        for value_set in values:
            i_expect_the_result_is_operator_than_number(_step_obj, operator, number, _stash=value_set)

    elif isinstance(values, dict):
        _step_obj.context.property_name = values.get('type', _step_obj.context.property_name)
        _step_obj.context.address = values.get('address', _step_obj.context.addresses)

        if isinstance(_step_obj.context.address, list) and len(_step_obj.context.address) == 1:
            _step_obj.context.address = _step_obj.context.address[0]

        i_expect_the_result_is_operator_than_number(_step_obj, operator, number, values.get('values', Null))

    elif isinstance(values, (int, str)):
        values = int(values)
        if operator in ('more', 'greater', 'bigger'):
            fail(values > number, '{} is not more than {}'.format(values, number))
        elif operator in ('more and equal', 'greater and equal', 'bigger and equal'):
            fail(values >= number, '{} is not more and equal than {}'.format(values, number))
        elif operator in ('less', 'lesser', 'smaller'):
            fail(values < number, '{} is not less than {}'.format(values, number))
        elif operator in ('less and equal', 'lesser and equal', 'smaller and equal'):
            fail(values <= number, '{} is not less and equal than {}'.format(values, number))
        elif operator in ('equal',):
            fail(values == number, '{} is not equal to {}'.format(values, number))
        else:
            raise TerraformComplianceNotImplemented('Invalid operator: {}'.format(operator))

    elif isinstance(values, Null):
        raise TerraformComplianceNotImplemented('Null/Empty value found on {}'.format(_step_obj.context.type))



# -*- coding: utf-8 -*-

from terraform_compliance.common.defaults import Defaults
from terraform_compliance.common.helper import (
    get_resource_address_list_from_stash,
    Null
)
from terraform_compliance.extensions.ext_radish_bdd import skip_step


def its_key_is_value(_step_obj, key, value, dict_value=None, address=Null):
    match = _step_obj.context.match
    
    orig_key = key
    if key == 'reference':
        if address is not Null:
            key = Defaults.r_mount_addr_ptr
        elif address is Null:
            key = Defaults.r_mount_addr_ptr_list

    found_list = []
    for obj in _step_obj.context.stash:
        object_key = obj.get('values', {})
        if isinstance(object_key, list):
            for el in object_key:
                if isinstance(el, dict):
                    filtered_key = match.get(el, key)
                    if isinstance(filtered_key, (str, int, bool)) and match.equals(filtered_key, value):
                        found_list.append(el)
        else:
            object_key = match.get(object_key, key, Null)

        if object_key is Null:
            object_key = match.get(obj, key, Null)
            if address is not Null and isinstance(object_key, dict) and match.contains(object_key, address):
                object_key = match.get(object_key, address, Null)

        if isinstance(object_key, str):
            if "[" in object_key:
                object_key = object_key.split('[')[0]

            if match.equals(object_key, value):
                found_list.append(obj)

        elif isinstance(object_key, (int, bool)) and match.equals(object_key, value):
            found_list.append(obj)

        elif isinstance(object_key, list):
            object_key = [str(v) for v in object_key]
            if match.contains(object_key, value):
                found_list.append(obj)

        elif isinstance(object_key, dict):
            candidate_value = match.get(object_key, value)
            if candidate_value is not None and (
                    dict_value is None or (
                    match.equals(candidate_value, dict_value))
            ):
                found_list.append(obj)

        elif object_key is None and match.equals('None', value):
            found_list.append(obj)

    if found_list != []:
        _step_obj.context.stash = found_list
        _step_obj.context.addresses = get_resource_address_list_from_stash(found_list)
    else:
        if object_key is Null:
            skip_step(_step_obj, message='Could not find {} in {}.'.format(key,
                                                                           ', '.join(_step_obj.context.addresses)))
        elif dict_value is None:
            skip_step(_step_obj, message='Can not find {} {} in {}.'.format(value, orig_key,
                                                                            ', '.join(_step_obj.context.addresses)))
        else:
            skip_step(_step_obj, message='Can not find {}={} {} in {}.'.format(value, dict_value, orig_key,
                                                                               ', '.join(_step_obj.context.addresses)))


def its_key_is_not_value(_step_obj, key, value, dict_value=None, address=Null):
    match = _step_obj.context.match
    
    orig_key = key
    if key == 'reference':
        if address is not Null:
            key = Defaults.r_mount_addr_ptr
        elif address is Null:
            key = Defaults.r_mount_addr_ptr_list

    found_list = []
    for obj in _step_obj.context.stash:
        object_key = obj.get(key, Null)

        if object_key is Null:
            object_key = obj.get('values', {})
            if isinstance(object_key, list):
                object_keys = []
                for object_key_element in object_key:
                    if not match.equals(match.get(object_key_element, key, Null), value):
                        object_keys.append(match.get(object_key_element, key, Null))

                object_key = [keys for keys in object_keys if keys is not Null]
            else:
                object_key = match.get(object_key, key, Null)

        if address is not Null and isinstance(object_key, dict) and match.contains(object_key, address):
            object_key = match.get(object_key, address, Null)

        if isinstance(object_key, str):
            if "[" in object_key:
                object_key = object_key.split('[')[0]

            if not match.equals(object_key, value):
                found_list.append(obj)

        elif isinstance(object_key, (int, bool)) and not match.equals(object_key, value):
            found_list.append(obj)

        elif isinstance(object_key, list) and not match.contains(object_key, value):
            found_list.append(obj)

        elif isinstance(object_key, dict):
            if not match.contains(object_key, value) or (dict_value is not None and not match.equals(str(match.get(object_key, value)), dict_value)):
                found_list.append(obj)

        elif object_key is None and not match.equals('None', value):
            found_list.append(obj)

    if found_list != []:
        _step_obj.context.stash = found_list
        _step_obj.context.addresses = get_resource_address_list_from_stash(found_list)
    else:
        if object_key is Null:
            skip_step(_step_obj, message='Could not find {} in {}.'.format(key,
                                                                     ', '.join(_step_obj.context.addresses)))
        elif dict_value is None:
            skip_step(_step_obj, message='Found {} {} in {}.'.format(value, orig_key,
                                                                     ', '.join(_step_obj.context.addresses)))
        else:
            skip_step(_step_obj, message='Found {}={} {} in {}.'.format(value, dict_value, orig_key,
                                                                        ', '.join(_step_obj.context.addresses)))



from terraform_compliance.common.error_handling import Error
from terraform_compliance.common.helper import Null

def it_must_be_in(_step_obj, haystack, cover=False):
    match = _step_obj.context.match

    values = [resource['values'] for resource in _step_obj.context.stash]
    for value in values:
        if not isinstance(value, (bool, int, float, str)):
            Error(_step_obj, 'Values are not in the correct type. (bool, str, int, float)')

    if 'haystack' not in _step_obj.context.in_step_variables:
        Error(_step_obj, 'Could not resolve haystack: {}'.format(haystack))       
    parsed_haystack = _step_obj.context.in_step_variables['haystack']

    if not cover and not match.issubset(values, parsed_haystack):
        Error(_step_obj, '{} does not cover selected resources.'.format(haystack))                                                                        

    if cover and not match.issubset(parsed_haystack, values):
        Error(_step_obj, 'Selected resources does not cover {}.'.format(haystack))                                                                        



# -*- coding: utf-8 -*-

from terraform_compliance.steps import property_match_list
from terraform_compliance.common.helper import (
    Null
)
from terraform_compliance.common.error_handling import Error


def property_is_enabled(_step_obj, something):
    match = _step_obj.context.match
    seek_key_in_dict = match.seek_key_in_dict

    for resource in _step_obj.context.stash:
        if isinstance(resource, dict):
            if match.contains(property_match_list, something):
                something = match.get(match.get(property_match_list, something), resource['type'], something)

            property_value = seek_key_in_dict(match.get(resource, 'values', {}), something)

            if len(property_value):
                property_value = property_value[0]

                if isinstance(property_value, dict):
                    property_value = match.get(property_value, something, Null)

            if not property_value:
                Error(_step_obj, 'Resource {} does not have {} property enabled '
                                 '({}={}).'.format(resource.get('address', "resource"),
                                                   something,
                                                   something,
                                                   property_value))
    return True



# -*- coding: utf-8 -*-

"""
    This module provides a class to represent a Step
"""

from __future__ import unicode_literals

import base64
import re

from radish.model import Model
from radish.exceptions import RadishError
from radish.terrain import world
from radish.stepregistry import StepRegistry
from radish.matcher import merge_step
from radish import utils


class Step(Model):
    """
        Represents a step
    """

    class State(object):
        """
            Represents the step state

            FIXME: for the python3 version this should be an Enum
        """

        UNTESTED = "untested"
        SKIPPED = "skipped"
        PASSED = "passed"
        FAILED = "failed"
        PENDING = "pending"

    def __init__(self, id, sentence, path, line, parent, runable, context_class=None):
        super(Step, self).__init__(id, None, sentence, path, line, parent)
        self.context_class = context_class
        self.table_header = None
        self.table_data = []
        self.table = []
        self.raw_text = []
        self.definition_func = None
        self.argument_match = None
        self.state = Step.State.UNTESTED
        self.failure = None
        self.runable = runable
        self.as_precondition = None
        self.as_background = None
        self.embeddings = []

    @property
    def context(self):
        """
            Returns the scenario context belonging to this step
        """
        return self.parent.context

    @property
    def expanded_sentence(self):
        """
            Returns the expanded sentence of this step

                * Expand constants
        """
        sentence = self.sentence
        for name, value in self.parent.constants:
            sentence = sentence.replace("${{{0}}}".format(name), value)
        return sentence

    @property
    def context_sensitive_sentence(self):
        """
        Return the context class sensitive
        step sentence.
        """
        sentence = self.expanded_sentence
        if self.context_class:
            return utils.str_lreplace(
                "and ",
                self.context_class.capitalize() + " ",
                sentence,
                flags=re.IGNORECASE,
                )

        return sentence

    @property
    def text(self):
        """
            Returns the additional text of this step as string
        """
        return "\n".join(self.raw_text)

    def _validate(self):
        """
            Checks if the step is valid to run or not
        """

        if not self.definition_func or not callable(self.definition_func):
            raise RadishError(
                "The step '{0}' does not have a step definition".format(self.sentence)
            )

    def run(self):
        """
            Runs the step.
        """
        if not self.runable:
            self.state = Step.State.UNTESTED
            return self.state

        self._validate()
        args, kwargs = self.argument_match.evaluate()

        try:
            if kwargs:
                self.definition_func(self, **kwargs)  # pylint: disable=not-callable
            else:
                self.definition_func(self, *args)  # pylint: disable=not-callable
        except Exception as e:  # pylint: disable=broad-except
            self.state = Step.State.FAILED
            self.failure = utils.Failure(e)
        else:
            if self.state is Step.State.SKIPPED:
                self.skip()
            elif self.failure is not None and self.state is Step.State.FAILED:
                self.state = Step.State.FAILED
            elif self.state is not Step.State.PENDING:
                self.state = Step.State.PASSED
        return self.state

    def debug(self):
        """
            Debugs the step
        """
        if not self.runable:
            self.state = Step.State.UNTESTED
            return self.state

        self._validate()
        args, kwargs = self.argument_match.evaluate()

        pdb = utils.get_debugger()

        try:
            pdb.runcall(self.definition_func, self, *args, **kwargs)
        except Exception as e:  # pylint: disable=broad-except
            self.state = Step.State.FAILED
            self.failure = utils.Failure(e)
        else:
            if self.state is Step.State.SKIPPED:
                self.skip()
            elif self.state is not Step.State.PENDING:
                self.state = Step.State.PASSED
        return self.state

    def skip(self):
        """
            Skips the step
        """
        self.state = Step.State.SKIPPED

    def pending(self):
        """
            Mark the step as pending

            The end report will remind
            about all pending steps.
        """
        self.state = Step.State.PENDING

    def behave_like(self, sentence):
        """
            Make step behave like another one

            :param string sentence: the sentence of the step to behave like
        """
        # check if this step has already failed from a previous behave_like call
        if self.state is Step.State.FAILED:
            return

        # create step according to given sentence
        new_step = Step(None, sentence, self.path, self.line, self.parent, True)
        merge_step(new_step, StepRegistry().steps)

        # run or debug step
        if world.config.debug_steps:
            new_step.debug()
        else:
            new_step.run()

        # re-raise exception if the failed
        if new_step.state is Step.State.FAILED:
            new_step.failure.exception.args = (
                "Step '{0}' failed: '{1}'".format(
                    sentence, new_step.failure.exception.message
                ),
            )
            raise new_step.failure.exception

    def embed(self, data, mime_type="text/plain", encode_data_to_base64=True):
        """
            embed data into step
                - step embedded data can be used for cucumber json reports
                - allow to attach text, html or images

        :param data: data attached to report
            data needs to be encoded in base64 for proper handling by tests reporting tools
            if not encoded in base64 encoded_data_to_base64 needs to be True (default value)
        :param mime_type: image/png, image/bmp, text/plain, text/html
            - mime types with special support by Cucumber reporting
            - other mime types will be handled with default handling
        :param encode_data_to_base64: encode data to base64 if True
        """
        if encode_data_to_base64:
            data = base64.standard_b64encode(data.encode()).decode()
        self.embeddings.append({"data": data, "mime_type": mime_type})
